import random

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.colors import ListedColormap

def standardize(arr, stats=None):
    """Standardize the input.

    Parameters
    ----------
    arr : numpy.ndarray
        An array to be standardized.
    stats : tuple of float, default None
        Average and standard deviation used to standardize the arr.
        If None, those of the arr are used.

    Returns
    -------
    arr_std : numpy.ndarray
        A standardized array
    stats : tuple of float
        Average and standard deviation used to standardize the input
        array.
    """
    if arr.ndim > 2:
        raise ValueError("'arr' must be 1D or 2D.")

    if stats is None:
        if arr.ndim == 1:
            stats = (arr.mean(), arr.std())
        else:
            stats = (arr.mean(axis=1), arr.std(axis=1))

    if arr.ndim == 1:
        arr_std = (arr - stats[0]) / stats[1]
    else:
        arr_std = (arr - stats[0].reshape(-1, 1)) / stats[1].reshape(-1, 1)

    return arr_std, stats


class PolynomialFeatures():
    """A class like sklearn.preprocessing.PolynomialFeatures."""
    def __init__(self, degree):
        """
        Parameters
        ----------
        degree : int
            degree of polynomial plane
        """
        self.__degree = degree

    def transform(self, r):
        """This is similar to fit_transform, but the returned array is
        transported."""
        n = (self.__degree + 2) * (self.__degree + 1) // 2 - 1
        arr = np.empty((n, r.shape[1]))

        k = 0
        for i in range(1, self.__degree+1):
            for j in range(i + 1):
                arr[k,:] = r[0,:]**(i - j) * r[1,:]**j
                k += 1
        return arr


class GuessInitRsp():
    """Class for making a guess of initial responsibility.

    Parameters
    ----------
    array : array-like
        image array
    n : ``int``
        Number of terraces
    threshold : ``float``
        When difference of values between adjacent pixels is below
        this threshold, those pixels are regared in the same terrace.
    seeds : array-like of ``tuple``
        Use this as initial locations of clustering, if given. Each tuple
        denotes an initial location. If number of tuples is less than ``n``,
        random initial locations are used for the rest.
    min_size : ``int``, default 10
        Minimum number of pixels of a terrace.

    Attributes
    ----------
    guess : ``numpy.ndarray``
        An array of suggested initial responsibility
    image : ``numpy.ndarray``
        An array of an image shown by show(). A value at each pixel denotes
        a terrace in which the pixel is assigned. -1 means unassigned.

    """
    def __init__(self, array, n, threshold, seeds=None, min_size=10):
        self.__array = np.asarray(array)
        self.seeds = []
        self.guess = self.__make_guess(n, threshold, seeds, min_size)
        self.image = self.__make_image()
        print(self)

    def __str__(self):
        msg = 'terrace  seeds  pixels\n' \
              '----------------------\n'
        for i in range(self.guess.shape[0]):
            msg += f'{i}: '\
                   f'({self.seeds[i][0]:.0f}, {self.seeds[i][1]:.0f}) '\
                   f'{self.guess[i,:,:].sum()}\n'
        msg += f'coverage {self.guess.sum()/self.__array.size*100:.0f}% ' \
               f'({self.guess.sum()}/{self.__array.size})'
        return msg

    def __make_guess(self, n, thresh, seeds, min_size):
        """
        1. select a point from self.__seeds or a random point that is not yet
           included in any clusters
        2. grow a cluster from it
        3. if the cluster has a point that is already included in an existing
           cluster, merge them
        4. repeat 1-3 until reaching the specified number of clusters
        """
        _seeds = list(reversed(seeds)) if seeds is not None else[]
        guess = np.zeros((n,) + self.__array.shape, dtype=bool)
        i = 0
        while i < n:
            seed = _seeds.pop() if _seeds else self.__random_pickup(guess)
            cluster = self.__grow_cluster(seed, thresh)
            if cluster.sum() < min_size:
                continue
            guess[i,:,:] = cluster
            self.seeds.append(seed)
            i += 1
        return guess

    def __grow_cluster(self, init_loc, thresh):
        """Grow a cluster from a point specified by ``init_loc``. Two
        points are regarded in the same cluster if difference of values
        of the two points is below ``self.__thresh``."""
        cluster = np.zeros_like(self.__array, dtype=bool)
        cluster[init_loc[0], init_loc[1]] = True
        pocket = [init_loc]
        adjacent = [(-1, 0), (1, 0), (0, -1), (0, 1)]
        m, n = self.__array.shape
        while pocket:
            pt = pocket.pop(0)
            neighbors_in_cluster = [
                    (pt[0] - i, pt[1] - j) for (i, j) in adjacent
                    if 0 <= pt[0] - i < m and 0 <= pt[1] - j < n and
                    not cluster[pt[0] - i, pt[1] - j] and
                    np.absolute(self.__array[pt[0], pt[1]]
                                - self.__array[pt[0] - i, pt[1] - j])
                    < thresh]
            for nbr in neighbors_in_cluster:
                pocket.append(nbr)
                cluster[nbr[0], nbr[1]] = True
        return cluster

    def __random_pickup(self, guess):
        """Return a pixel that is not yet included in any clusters."""
        already_clustered = guess.sum(axis=0)
        while True:
            p1 = random.randint(0, guess.shape[1] - 1)
            p2 = random.randint(0, guess.shape[2] - 1)
            if not already_clustered[p1, p2]:
                return (p1, p2)

    def __make_image(self):
        image = np.full((self.guess.shape[1], self.guess.shape[2]), -1)
        for i in range(self.guess.shape[0]):
            image += self.guess[i,:,:] * (i + 1)
        return image

    def show(self, **kwargs):
        """Show an image of suggested initial responsibility.

        Parameters
        ----------
        **kwargs :
            keyword parameters to be passed to matplotlib.pyplot.imshow()
        """
        n = self.guess.shape[0]
        cmap_name = 'tab20' if n > 10 else 'tab10'
        clrs = ((1., 1., 1.),) + get_cmap(cmap_name).colors
        cmap = ListedColormap(np.delete(clrs, slice(n+1, None), axis=0))
        self.image = self.__make_image()
        plt.imshow(self.image, cmap = cmap, origin = 'lower', interpolation = None,
                   resample=False, vmin=-1.5, vmax=n-0.5, **kwargs)
        plt.colorbar(label='index', ticks=list(range(-1, n)))

        # The __test attribute is used not to block test
        block = not hasattr(self, f'_{self.__class__.__name__}__test')
        plt.show(block=block)





__all__ = ['Norm', 'Cauchy', 'T', 'VonMises']

from abc import ABCMeta, abstractmethod

import numpy as np
import numexpr as ne
import scipy.special


class Distribution(metaclass=ABCMeta):
    """A base class to construct specific distribution classes.

    All distributions take ``loc`` and ``scale`` as keyword parameters to
    adjust the location and scale of the distribution, just as
    `scipy.stat
    <https://docs.scipy.org/doc/scipy/reference/tutorial/stats.html#shifting-and-scaling>`_,
    unless otherwise noted.

    Optimized parameters are stored in attributes of the same name.
    """
    def __init__(self, loc=None, scale=None):
        """
        Parameters
        ----------
        loc : array-like, default None
            Initial guess of the location parameter.
        scale : array-like, default None
            Initial guess of the scale parameter.
        """
        self.loc = None if loc is None else np.array(loc)
        self.scale = None if scale is None else np.array(scale)

    def __str__(self):
        np.set_printoptions(precision=3)
        return f'centers: {self.loc}\nscales:  {self.scale}'

    def scaler(self, image_scale, backward=False):
        """Apply scaling used to standardize an input image to
        distribution parameters.

        Parameters
        ----------
        image_scale : tuple of float
            Values used to standardize an input image, mean and std.
        backward : bool
            If true, apply the reverse process to distribution parameter
            in order to restore the original scale of an input image.
        """
        (mean, std) = image_scale
        if backward:
            self.loc = self.loc * std + mean
            self.scale = self.scale * std
            return
        if self.loc is not None:
            self.loc = (self.loc - mean) / std
        if self.scale is not None:
            self.scale = self.scale / std

    @abstractmethod
    def pdf(self, t_subtracted):
        """Probability distribution function.

        Parameters
        ----------
        t_subtracted : numpy.ndarray
            One-dimensionalized image subtracted by a model plane.

        Returns
        -------
        numpy.ndarray, 2D
            The first dimension is class, and the second dimension is position.
        """

    def initial_step(self, t_subtracted, gamma):
        """Calculate distribution parameters in the initial step.
        """
        self.pi = self.mixing_coef(gamma)
        if self.loc is None:
            self.loc = self.center(t_subtracted, gamma)
        if self.scale is None:
            self.scale = self.std(t_subtracted, gamma, self.loc)

    def retain(self):
        """Retain the present distribution parameters so that they can be
        reverted later."""
        self.__loc, self.__scale = self.loc, self.scale

    def revert(self):
        """Revert the distribution parameters to those retained beforehand.
        """
        self.loc, self.scale = self.__loc, self.__scale

    @abstractmethod
    def grad(self, t_subtracted, gamma):
        """Calculate derivative of log likelihood with respect to distribution
        parameters.

        Parameters
        ----------
        t_subtracted : numpy.ndarray, 1D
            One-dimensionalized image subtracted by a model plane.
        gamma : numpy.ndarray, 2D
            Two-dimensionalized responsibility. The first and second
            dimensions are class and position, responsibility.

        Returns
        -------
        tuple of numpy.ndarray
            By summing an ndarray of this list along axis 1, derivative
            of the log likelihood is obtained. The ndarrays are returned
            without being summed because the first element (grad_loc_like)
            is also used to calculate derivative with respect to the plane
            coefficients.
        """

    def update_ga(self, new):
        """Update distribution parameters in the M-step.

        Parameters
        ----------
        new : numpy.ndarray
            The size is product of numbers of classes and parameters.
        """
        if isinstance(self.loc, float):
            self.loc, self.scale = new[0], new[1]
        else:
            K = self.loc.size
            self.loc, self.scale = new[:K], new[K:K * 2]

    def x0(self):
        """Create a part of x0 used in scipy.optimize.minimize"""
        return np.concatenate([self.loc, self.scale], axis=None)

    def lb(self):
        """Create a part of lb used in scipy.optimize.Bounds"""
        return np.concatenate([np.full_like(self.loc, -np.inf),
                               np.full_like(self.scale, np.finfo(float).eps)],
                              axis=None)

    def center(self, t, gamma):
        """Calculate center values of a distribution.

        Parameters
        ----------
        t : numpy.ndarray, 1D
            One-dimensionalized image, or one subtracted by a model plane.
        gamma : numpy.ndarray, 2D
            Two-dimensionalized responsibility. The first and second
            dimensions are class and position, responsibility.

        Returns
        -------
        numpy.ndarray
            The center values of each class.
        """
        return np.array([np.average(t, weights=gamma[k])
                         for k in range(gamma.shape[0])])

    def std(self, t, gamma, loc):
        """Calculate standard deviation of a distribution.

        Parameters
        ----------
        t : numpy.ndarray
            One-dimensionalized image, or one subtracted by a model plane.
        gamma : numpy.ndarray
            Two-dimensionalized responsibility.
        loc : numpy.ndarray
            Center values.

        Returns
        -------
        numpy.ndarray
            Standard deviation of each class.
        """
        return np.array([
            0 if (n := np.sum(gamma[k])) == 0
            else np.sqrt(np.sum(np.square(t - loc[k]) * gamma[k]) / n)
            for k in range(gamma.shape[0])])

    def mixing_coef(self, gamma):
        """Calculate mixing coefficients.

        Parameters
        ----------
        gamma : numpy.ndarray
            Two-dimensionalized responsibility.

        Returns
        -------
        numpy.ndarray
            The mixing coefficients of each class.
        """
        return (N_k := gamma.sum(axis=1)) / N_k.sum()

    def fullpdf(self, t_subtracted):
        """Probability density function multiplied by mixing coefficient.

        Parameters
        ----------
        t_subtracted : numpy.ndarray
            One-dimensionalized image subtracted by a model plane.

        Returns
        -------
        numpy.ndarray
            The first dimension is class, and the second dimension is
            position.
        """
        return self.pi.reshape(-1, 1) * self.pdf(t_subtracted)


class Norm(Distribution):
    r"""Class for normal distribution,

    .. math::
       p(x|\mu, \sigma) = \dfrac{1}{\sqrt{2\pi\sigma^2}}
           \exp\left\{-\dfrac{(x-\mu)^2}{2\sigma^2}\right\}

    where :math:`\mu` and :math:`\sigma` are the location and scale
    parameter, respectively.

    Parameters
    ----------
    loc : array-like, default ``None``
        Initial guess of the location parameter. By default, this is
        automatically calculated.
    scale : array-like, default ``None``
        Initial guess of the scale parameter. By default, this is
        automatically calculated.

    Attributes
    ----------
    loc, scale : ``numpy.ndarray``
        Optimized parameters.
    """
    def pdf(self, t_subtracted):
        p = np.pi * 2
        t = t_subtracted.reshape(1, -1)
        s2, l = (self.scale**2).reshape(-1, 1), self.loc.reshape(-1, 1)
        return ne.evaluate('exp(-(t - l)**2 / (2 * s2)) / sqrt(p * s2)')

    def update_quick(self, t_subtracted, gamma):
        """Calculate distribution parameters in the M-step using the
        quick method.

        Parameters
        ----------
        t_subtracted : numpy.ndarray, 1D
            One-dimensionalized image subtracted by a model plane.
        gamma : numpy.ndarray, 2D
            Two-dimensionalized responsibility. The first and second
            dimensions are class and position, responsibility.
        """
        self.weight = gamma
        self.retain()
        self.loc = self.center(t_subtracted, gamma)
        self.scale = self.std(t_subtracted, gamma, self.loc)
        self.pi = self.mixing_coef(gamma)

    def grad(self, t_subtracted, gamma):
        # Derivative is obtained by summing the following arrays in axis 1
        # (position). But an array before summation (grad_loc_like) is used
        # for calculating derivative with respect to plane coefficients.
        # Therefore, arrays are returned without the summation.
        p = np.pi * 2
        pdf = self.pdf(t_subtracted)
        t = t_subtracted.reshape(1, -1)
        s, l = self.scale.reshape(-1, 1), self.loc.reshape(-1, 1)
        grad_loc_like = ne.evaluate('gamma * (t - l) / s**2')
        grad_scale_like = ne.evaluate('gamma * (t**2 - 2 * t * l + l**2 - s**2) / s**3')
        return (grad_loc_like, grad_scale_like)


class Cauchy(Distribution):
    r"""Class for Cauchy distribution.

    .. math::
       p(x|\mu, \sigma) = \dfrac{1}{\pi}
                          \dfrac{\sigma}{(x-\mu)^2+\sigma^2}

    where :math:`\mu` and :math:`\sigma` are the location and scale
    parameter, respectively.

    Parameters
    ----------
    loc : array-like, default ``None``
        Initial guess of the location parameter. By default, this is
        automatically calculated.
    scale : array-like, default ``None``
        Initial guess of the scale parameter. By default, this is
        automatically calculated.

    Attributes
    ----------
    loc, scale : ``numpy.ndarray``
        Optimized parameters.
    """
    def pdf(self, t_subtracted):
        p = np.pi
        t = t_subtracted.reshape(1, -1)
        s, l = self.scale.reshape(-1, 1), self.loc.reshape(-1, 1)
        return ne.evaluate('s / (((t-l)**2 + s**2) * p)')

    def update_quick(self, t_subtracted, gamma):
        """Calculate distribution parameters in the M-step using the quick
        method.

        Parameters
        ----------
        t_subtracted : numpy.ndarray, 1D
            One-dimensionalized image subtracted by a model plane.
        gamma : numpy.ndarray, 2D
            Two-dimensionalized responsibility. The first and second
            dimensions are class and position, responsibility.
        """
        p = np.pi
        t = t_subtracted.reshape(1, -1)
        s, l = self.scale.reshape(-1, 1), self.loc.reshape(-1, 1)
        self.weight = ne.evaluate('gamma / (((t-l)**2 + s**2) * p)')
        gamma_tilde = self.weight * self.scale.reshape(-1, 1)
        self.retain()
        self.loc = self.center(t_subtracted, gamma_tilde)
        self.scale = self.__scaling_param(gamma, gamma_tilde)
        self.pi = self.mixing_coef(gamma)

    def __scaling_param(self, gamma, gamma_tilde):
        numerator = gamma.sum(axis=1)
        denominator = gamma_tilde.sum(axis=1) * 2 * np.pi
        return np.divide(numerator, denominator, where=denominator!=0)

    def grad(self, t_subtracted, gamma):
        p = np.pi * 2
        pdf = self.pdf(t_subtracted)
        t = t_subtracted.reshape(1, -1)
        s, l = self.scale.reshape(-1, 1), self.loc.reshape(-1, 1)
        # Derivative is obtained by summing the following arrays in axis 1
        # (position). But an array before summation (grad_loc_like) is used
        # for calculating derivative with respect to plane coefficients.
        # Therefore, arrays are returned without the summation.
        grad_loc_like = ne.evaluate('gamma * pdf * p / s * (t - l)')
        grad_scale_like = ne.evaluate('gamma * (1 / s - pdf * p)')
        return (grad_loc_like, grad_scale_like)


class T(Distribution):
    r"""Class for Student's t distribution,

    .. math::
       p(x|\mu, \sigma, \mu) =
            \dfrac{\Gamma((\nu+1)/2)}{\Gamma(\nu/2)\sigma\sqrt{\pi\nu}}
            \left\{1+\dfrac{1}{\nu}
                   \left(\dfrac{x-\mu}{\sigma}\right)^2\right\}
            ^{-(\nu+1)/2}

    where :math:`\mu`, :math:`\sigma`, and :math:`\nu` are the location,
    scale, and degree of freedom parameter, respectively.

    Parameters
    ----------
    loc : array-like, default ``None``
        Initial guess of the location parameter. By default, this is
        automatically calculated.
    scale : array-like, default ``None``
        Initial guess of the scale parameter. By default, this is
        automatically calculated.
    df : array-like, default ``None``
        Initial guess of the degree of freedom. By default, this is
        ``numpy.ones_like(loc)``.

    Attributes
    ----------
    loc, scale, df : ``numpy.ndarray``
        Optimized parameters.
    """
    def __init__(self, loc=None, scale=None, df=None):
        super().__init__(loc=loc, scale=scale)
        self.df = None if df is None else np.array(df)

    def __str__(self):
        return super().__str__() + f'\ndegree of freedom: {self.df}'

    def pdf(self, t_subtracted):
        u = (1 / (self.scale**2 * self.df)).reshape(-1, 1)
        v = self.df.reshape(-1, 1)
        return np.sqrt(u) / scipy.special.beta(v * 0.5, 0.5) \
            * (1 + u * (t_subtracted.reshape(1, -1) \
            - self.loc.reshape(-1, 1))**2)**(-0.5*v - 0.5)

    def initial_step(self, t_subtracted, gamma):
        super().initial_step(t_subtracted, gamma)
        if self.df is None:
            self.df = np.ones_like(self.loc)

    def retain(self):
        super().retain()
        self.__df = self.df

    def revert(self):
        super().revert()
        self.df = self.__df

    def grad(self, t_subtracted, gamma):
        x = t_subtracted.reshape(1, -1) - self.loc.reshape(-1, 1)
        y = 1 / self.scale
        u = (x * y.reshape(-1, 1))**2 + self.df.reshape(-1, 1)
        v = (self.df.reshape(-1, 1) + 1) * np.reciprocal(u)
        w = 1 + scipy.special.psi((self.df + 1) * 0.5) \
            - scipy.special.psi(self.df * 0.5) + np.log(self.df)
        # Derivative is obtained by summing the following arrays in axis 1
        # (position). But an array before summation (grad_loc_like) is used
        # for calculating derivative with respect to plane coefficients.
        # Therefore, arrays are returned without the summation.
        grad_loc_like = gamma * x * v * (y ** 2).reshape(-1, 1)
        grad_scale_like = gamma * (1 - v) * (self.df * y).reshape(-1, 1)
        grad_df_like = gamma * 0.5 * (w.reshape(-1, 1) - np.log(u) - v)
        return(grad_loc_like, grad_scale_like, grad_df_like)

    def update_ga(self, new):
        super().update_ga(new)
        self.df = new[self.df.size*2:]

    def x0(self):
        return np.concatenate([self.loc, self.scale, self.df], axis=None)

    def lb(self):
        inf = np.full_like(self.loc, -np.inf)
        epsilon = np.full_like(self.scale, np.finfo(float).eps)
        return np.concatenate([inf, epsilon, epsilon], axis=None)


class VonMises(Distribution):
    r"""Class for von Mises distribution.

    .. note::

        This class is intended as a prior distribution for the location
        parameter :math:`\mu` of either :class:`sati.distributions.Norm`,
        :class:`sati.distributions.Cauchy`, or
        :class:`sati.distributions.T`.

    .. math::
       p(\mu|c_0, \varphi_0, \kappa) = \dfrac{1}{2\pi I_0(\kappa)}
           \exp\left\{
               \kappa\cos\left(\dfrac{2\pi\mu}{c_0}-\varphi_0\right)
           \right\}

    where :math:`\varphi_0`, :math:`c_0`, and :math:`\kappa` are the
    location, scale, and concentration parameters, respectively.
    :math:`I_0` is the modified Bessel function of the first kind
    at order 0.

    This is slightly different from the standard form of von Mises
    distribution to estimate a unit height of steps.

    Parameters
    ----------
    loc : ``float``, default 0
        Initial guess of the location parameter :math:`\varphi_0`.
    scale : ``float``
        Initial guess of the scale parameter :math:`c_0`. This takes
        an initial guess of a unit height of steps.
    kappa : array-like, default 1.0
        The concentration parameter. The length of ``kappa`` must be the
        same as number of terraces. To exclude a terrace from the MAP
        estimation, set the corresponding element of ``kappa`` to 0.

    Attributes
    ----------
    loc, scale  : ``float``
        Optimized parameters.
    """
    def __init__(self, *, loc=0., scale, kappa=[1.]):
        self.loc, self.scale, self.__kappa = loc, scale, np.asarray(kappa)

    def __str__(self):
        return f'spacing: {self.scale:.3f}'

    def pdf(self, mu):
        """Probability distribution function.

        Parameters
        ----------
        mu : ``numpy.ndarray``
            The ``loc`` attribute of either :class:`sati.distributions.Norm`,
            :class:`sati.distributions.Cauchy`, or
            :class:`sati.distributions.T` is expected.

        Returns
        -------
        ``numpy.ndarray``
            Values at each class (terrace)
        """
        return np.exp(self.__kappa * np.cos(2 * np.pi * mu / self.scale
                                            - self.loc)
                     ) / (2 * np.pi * scipy.special.i0(self.__kappa))

    def loglikelihood_at_pixel(self, mu):
        """
        Parameters
        ----------
        mu : ``numpy.ndarray``
            The ``loc`` attribute of either :class:`sati.distributions.Norm`,
            :class:`sati.distributions.Cauchy`, or
            :class:`sati.distributions.T` is expected.
        """
        return (self.__kappa * np.cos(2 * np.pi * mu / self.scale - self.loc)
                - self._logi0(self.__kappa)).sum()

    def grad(self, mu, N):
        """
        Parameters
        ----------
        mu : ``numpy.ndarray``
            The ``loc`` attribute of either :class:`sati.distributions.Norm`,
            :class:`sati.distributions.Cauchy`, or
            :class:`sati.distributions.T` is expected.
        N : int
            Size in position.
        """
        u = self.__kappa * np.sin(2 * np.pi * mu / self.scale - self.loc)
        grad_mu = -2 * np.pi * N / self.scale * u
        grad_loc = N * u.sum()
        grad_scale = 2 * np.pi * N / (self.scale * self.scale) * (mu * u).sum()
        return (grad_mu, grad_loc, grad_scale)

    def scaler(self, image_scale, backward=False):
        (_, std) = image_scale
        self.scale = self.scale * std if backward else self.scale / std

    @classmethod
    def _logi0(cls, x):
        """This function returns a value of np.log(2*np.pi*np.i0(x)).
        When x >~ 800, scipy.special.i0(x) exceeds the limit of float64
        although np.log(scipy.special.i0(x)) is still not so large.
        To avoid this issue, np.log(np.i0(x)) is directly calculated
        without calculating np.i0(x). This is a logarithmic version of
        np.lib.function_base._i0_2(x), which is
        exp(x) * _chbevl(32.0/x - 2.0, _i0B) / sqrt(x)
        """
        return np.piecewise(x, [x < 700.], [cls.__logi0_1, cls.__logi0_2])

    @staticmethod
    def __logi0_1(x):
        return np.log(2 * np.pi * scipy.special.i0(x))

    @classmethod
    def __logi0_2(cls, x):
        return np.log(2 * np.pi) + x \
                + np.log(cls.__chbevl(32.0 / x - 2.0, cls.__i0B)) \
                - 0.5 * np.log(x)

    @staticmethod
    def __chbevl(x, vals):
        """This is a copy of np.lib.function_base._chbevl"""
        b0, b1 = vals[0], 0.0

        for i in range(1, len(vals)):
            b2 = b1
            b1 = b0
            b0 = x*b1 - b2 + vals[i]

        return 0.5 * (b0 - b2)

    # This is a copy of np.lib.function_base._i0B
    __i0B = (
        -7.23318048787475395456E-18,
        -4.83050448594418207126E-18,
        4.46562142029675999901E-17,
        3.46122286769746109310E-17,
        -2.82762398051658348494E-16,
        -3.42548561967721913462E-16,
        1.77256013305652638360E-15,
        3.81168066935262242075E-15,
        -9.55484669882830764870E-15,
        -4.15056934728722208663E-14,
        1.54008621752140982691E-14,
        3.85277838274214270114E-13,
        7.18012445138366623367E-13,
        -1.79417853150680611778E-12,
        -1.32158118404477131188E-11,
        -3.14991652796324136454E-11,
        1.18891471078464383424E-11,
        4.94060238822496958910E-10,
        3.39623202570838634515E-9,
        2.26666899049817806459E-8,
        2.04891858946906374183E-7,
        2.89137052083475648297E-6,
        6.88975834691682398426E-5,
        3.36911647825569408990E-3,
        8.04490411014108831608E-1
    )



__all__ = ['Poly', 'Decay']

from abc import ABCMeta, abstractmethod

import numexpr as ne
import numpy as np
from scipy import linalg

import sati.preprocessing


class Plane(metaclass=ABCMeta):
    """A base class to construct specific plane classes."""
    @abstractmethod
    def setup(self, shape, roi):
        """Set up variables of a model plane.

        Parameters
        ----------
        shape : tuple of int
            Shape of an input image.
        roi : ndarray
            A 2D array indicating an region of interst.
        """

    @abstractmethod
    def update_ga(self, step):
        """Update the decay coefficients and a model plane."""

    @abstractmethod
    def x0(self):
        """Create a part of x0 used in scipy.optimize.minimize"""

    @abstractmethod
    def lb(self):
        """Create a part of lb used in scipy.optimize.Bounds"""

    @abstractmethod
    def retain(self):
        """Retain the present coefficients and plane so that they can be
        reverted later.
        """

    @abstractmethod
    def revert(self):
        """Revert the coefficients and plane to the retained ones."""

    @abstractmethod
    def grad(self, u):
        """Calculate derivative of log likelihood with respect to plane
        coefficients.

        Parameters
        ----------
        u : numpy.ndarray
            A part common to the derivative of log likelihood with respect
            to mu (center). The first and second dimensions are class and
            position, responsibility.

        Returns
        -------
        numpy.ndarray
            Derivative of log likelihood with respect to plane coefficients.
        """

    @abstractmethod
    def finalize(self):
        """Calculate a full plane."""


class Poly(Plane):
    r"""Class for a polynomial plane.

    .. math::
       f(n) = w_0 + w_1x_n + w_2y_n + w_3x_n^2 + w_4x_ny_n + w_5y_n^2 + \dots

    Parameters
    ----------
    degree : ``int``, default 1
        Degree of the polynomial plane.
    coef : ``numpy.ndarray``, default ``None``
        Initial guess of :math:`w_i`. By default, this is automatically
        caluculated from an input image and an initial guess of the
        responsibility.

    Attributes
    ----------
    coef, plane : ``numpy.ndarray``
        Optimized :math:`w_i` and :math:`f(n)`, responsibility.
    """
    def __init__(self, degree=1, coef=None):
        self.__degree = degree
        self.coef = coef

    def setup(self, shape, roi):
        self.__shape = shape
        self.__roi = roi

        # One-dimensionalized positions.
        x = np.tile(range(shape[1]), shape[0])
        y = np.arange(shape[0]).repeat(shape[1])

        # Separate the bias term from the others for standardization later.
        poly = sati.preprocessing.PolynomialFeatures(self.__degree)
        phi_poly = poly.transform(np.vstack((x, y)))
        phi_bias = np.ones(phi_poly.shape[1])

        # Standardize the polynomial terms and add the bias term after
        # the standardization. The first and second dimension of self.__phi
        # is degree and position, respectively. When an ROI is used, it is
        # applied before standardization.
        if roi is None:
            phi_poly_std, _ = sati.preprocessing.standardize(phi_poly)
            self.__phi = np.vstack((phi_bias, phi_poly_std))
        else:
            phi_poly_roi = np.compress(roi.reshape(-1) > 0, phi_poly, axis=1)
            phi_bias_roi = np.ones(phi_poly_roi.shape[1])
            phi_poly_roi_std, scaler = \
                sati.preprocessing.standardize(phi_poly_roi)
            phi_poly_std, _ = sati.preprocessing.standardize(phi_poly, scaler)
            self.__phi = np.vstack((phi_bias_roi, phi_poly_roi_std))
            self.__phi_full = np.vstack((phi_bias, phi_poly_std))

        if self.coef is not None:
            # Expand or truncate if necessary
            ndegree = self.__phi.shape[0]
            if self.coef.size < ndegree:
                self.coef = np.pad(self.coef, [0, ndegree - self.coef.size],
                                   'constant')
            else:
                self.coef = self.coef[:ndegree]

    def initial_step(self, t, gamma, mu):
        """Calculate initial values required in the initial step of the
        main loop.

        Unless plane coefficients are given, calculate those for each
        class from the initial responsibility and the initial center
        values, and use the average weighted by the initial responsibility
        as the initial coefficients.
        """
        if self.coef is None:
            nclass = gamma.shape[0]
            A = [(gamma[k].reshape(1, -1) * self.__phi) @ self.__phi.T
                    for k in range(nclass)]
            b = [self.__phi @ (gamma[k] * (t - mu[k])) for k in range(nclass)]
            x = [linalg.solve(A[k], b[k]) for k in range(nclass)]
            self.coef = np.average(x, axis=0, weights=gamma.sum(axis=1))
        self.plane = self.__calc_plane()

    def update_quick(self, t, weight, mu):
        """Calculate the plane coefficients and update a model plane with
        the coefficients. This is used in the quick method.
        """
        A = (weight.sum(axis=0).reshape(1, -1) * self.__phi) @ self.__phi.T
        b = self.__phi @ (t * weight.sum(axis=0) - mu @ weight)
        self.retain()
        self.coef = linalg.solve(A, b)
        self.plane = self.__calc_plane()

    def update_ga(self, new):
        self.coef = new
        self.plane = self.__calc_plane()

    def x0(self):
        return self.coef

    def lb(self):
        return np.full_like(self.coef, -np.inf)

    def retain(self):
        self.__coef, self.__plane = self.coef, self.plane

    def revert(self):
        self.coef, self.plane = self.__coef, self.__plane

    def grad(self, u):
        return self.__phi @ u.sum(axis=0)

    def finalize(self, std):
        if self.__roi is None:
            self.plane = self.plane.reshape(self.__shape) * std
        else:
            self.plane = np.dot(self.__phi_full.T, self.coef * std) \
                           .reshape(self.__shape)

    def __calc_plane(self):
        return self.coef @ self.__phi


class Decay(Plane):
    r"""Class for a decay plane.

    .. math::
       f(n) = \sum_i A_i\exp(-n/\tau_i)
       \quad\mathrm{or}\quad
       f(n) = \sum_i A_i\ln(n+\tau_i)

    Parameters
    ----------
    tau : ``float`` or array-like
        Initial guess of :math:`\tau_i` in the unit of pixels.
    coef : ``float`` or array-like, default 1
        Initial guess of :math:`A_i`
    kind : {'exp', 'log'}, default 'log'
        Decay function.
    orgdrct : ``str``, default 'lbx'
        Origin and direction of decay specifying chronological order of
        pixels in an image. The origin is the starting point of scan and
        the direction is that of fast scan. The origin is described by
        either 'lb', 'rb', 'lt', or 'rt', denoting left bottom, right
        bottom, left top, and right bottom, respectively. The direction
        is described by either 'x' or 'y'.

        .. image:: orgdrct.svg

        The characters are case-insensitive and have no particular order.

    Attributes
    ----------
    tau, coef, plane : ``numpy.ndarray``
        Optimized :math:`\tau_i`, :math:`A_i` and :math:`f(n)`,
        responsibility.
    orgdrct : ``str``
        Origin and direction of decay.
    """
    def __init__(self, *, tau, coef=None, kind='log', orgdrct='lbx'):
        self.tau = np.array(tau)
        if np.any(self.tau <= 0):
            raise ValueError("The 'tau' must be positive.")

        self.coef = None if coef is None else np.array(coef)

        if kind not in ('exp', 'log'):
            raise ValueError("The 'kind' argument must be 'exp' or 'log'.")
        self.__kind = kind

        if any([(c not in 'lrtbxy') for c in orgdrct.lower()]):
            raise ValueError('Unknown flag is included in orgdrct')
        # self.orgdrct is used in not this class but sati.Model to reorder
        # image, responsibility, and roi (if any)
        self.orgdrct = orgdrct.lower()

    def __str__(self):
        np.set_printoptions(precision=3)
        return f'{self.__kind} decay (tau): {self.tau} (pixel)\n' \
               f'          (A): {self.coef}'

    def setup(self, shape, scale, roi):
        self.__shape = shape
        self.__scale = scale
        self.__roi = roi

        self.__index = np.linspace(0, 1, num=np.prod(shape), endpoint=False)
        if roi is not None:
            self.__index = np.compress(roi.reshape(-1) > 0, self.__index)
        self.plane = np.empty_like(self.__index)

        if self.__kind == 'exp':
            self.__beta = -np.prod(shape) / np.array(self.tau)
        else:
            # use beta**2 instead of tau to guarantee that log takes
            # a positive number
            self.__beta = np.sqrt(np.array(self.tau) / np.prod(shape))

        self.coef = np.ones_like(self.__beta) if self.coef is None \
                else self.coef / self.__scale[1]

        self.grad = {'exp': self._grad_exp,
                     'log': self._grad_log}[self.__kind]
        self.__calc_plane = {'exp': self._calc_plane_exp,
                             'log': self._calc_plane_log}[self.__kind]

    def update_ga(self, new):
        self.__beta, self.coef = new[:self.__beta.size], new[self.__beta.size:]
        self.plane = self.__calc_plane()

    def x0(self):
        return np.concatenate([self.__beta, self.coef], axis=None)

    def lb(self):
        inf = np.full_like(self.coef, -np.inf)
        return np.concatenate([inf, inf], axis=None)

    def retain(self):
        self.__beta_ = self.__beta
        self.__coef, self.__plane = self.coef, self.plane

    def revert(self):
        self.__beta = self.__beta_
        self.coef, self.plane = self.__coef, self.__plane

    def grad(self):
        """This function is assigned to either _grad_exp() or _grad_log()
        in setup(). The names of functions to be assigned start with not __
        but _ for pickling."""

    def _grad_exp(self, u):
        i, b = self.__index.reshape(1, -1), self.__beta.reshape(-1, 1)
        v = u.sum(axis=0).reshape(1, -1) * ne.evaluate('exp(b * i)')
        grad__beta = ne.evaluate('sum(i * v, axis=1)') * self.coef
        grad_coef = ne.evaluate('sum(v, axis=1)')
        return np.concatenate([grad__beta, grad_coef])

    def _grad_log(self, u):
        i, b = self.__index.reshape(1, -1), self.__beta.reshape(-1, 1)
        v, w = u.sum(axis=0).reshape(1, -1), ne.evaluate('i + b**2')
        grad__beta = ne.evaluate('sum(v / w, axis=1)') * 2 * self.coef * self.__beta
        grad_coef = ne.evaluate('sum(v * log(w), axis=1)')
        return np.concatenate([grad__beta, grad_coef])

    def finalize(self, std):
        if self.__roi is None:
            self.plane = self.plane.reshape(self.__shape) * std
        else:
            index = np.linspace(0, 1, num=np.prod(self.__shape),
                                endpoint=False)
            if self.__kind == 'exp':
                self.plane = np.dot(np.exp(self.__beta.reshape(1, -1)
                                           * index.reshape(-1, 1)),
                                    self.coef * std).reshape(self.__shape)
            else:
                self.plane = np.dot(np.log(index.reshape(-1, 1)
                                           + (self.__beta**2).reshape(1, -1)),
                                    self.coef * std).reshape(self.__shape)

        if self.__kind == 'exp':
            self.tau = -np.prod(self.__shape) / self.__beta
        else:
            self.tau = self.__beta**2 * np.prod(self.__shape)
        self.coef *= self.__scale[1]

    def __calc_plane(self):
        """This function is assigned to either _calc_plane_exp() or
        _calc_plane_log() in setup(). The names of functions to be
        assigned start with not __ but _ for pickling."""

    def _calc_plane_exp(self):
        i, b = self.__index.reshape(-1, 1), self.__beta.reshape(1, -1)
        return np.dot(ne.evaluate('exp(b * i)'), self.coef)

    def _calc_plane_log(self):
        i, b = self.__index.reshape(-1, 1), self.__beta.reshape(1, -1)
        return np.dot(ne.evaluate('log(i + b**2)'), self.coef)



import numpy as np
import scipy.optimize


"""The default of optimizer options. ``reduction`` is used not in
``mimimize()`` but in the loop calling ``minimize()``. It's included
here for the documentation purpose."""
options_default = {
        'ftol': 1e-7, 'maxiter': 256, 'lr': 1e-3, 'betas': [0.9, 0.999],
        'epsilon': 1e-8, 'reduction': 0.5}


def minimize(fun, x0, *, method='adam', jac,
        options={'ftol': 1e-7, 'maxiter': 256, 'lr': 1e-3,
                 'betas': [0.9, 0.999], 'epsilon': 1e-8, 'reduction': 0.5}):
    """Minimization of scalar function of one or more variables.

    The parameters are similar to those of `scipy.optimize.minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize>`_.

    Parameters
    ----------
    fun : callable
        The objective function to be minimized: ``fun(x) -> float``
        where ``x`` is an array with shape (n,), where ``n`` is the number
        of independent variables.
    x0 : ``numpy.ndarray``
        Initial guess. Array of real elements of size (n,).
    method : {'adam'}, default 'adam'
        Optimizer.
    jac : callable
        A function that returns the gradient vector:
        ``jac(x) -> array_like, shape (n,)``
        where ``x`` is an array with shape (n,).
    options : ``dict``
        Optimizer options.

        * betas : ``tuple`` of ``float``
        * epsilon : ``float``

        These are parameters defined in the reference [1]_.

        * lr : ``float``, the learning rate.
        * ftol : ``float``, the fractional tolerance to stop the iteration.
        * maxiter : ``int``, the maximum number of iterations.
        * reduction : ``float``, this value is multiplied to ``lr`` when the objective function is smaller than that of the last step in the main EM loop.


    Returns
    -------
    scipy.optimize.OptimizeResult
        The optimization result. The results are stored in the attributes of
        ``x``, ``success``, ``fun``, ``jac``, and ``nit``. See
        `scipy.optimize.OptimizeResult <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html#scipy.optimize.OptimizeResult>`_ for more details.

    References
    ---=------
    .. [1] Diederik P. Kingma, Jimmy Ba, "Adam: A Method for Stochastic
       Optimization", arXiv:1412.6980
    """
    if type(options) == dict and options != options_default:
        options_ = options
        options = dict(options_default.items())
        options.update(options_)

    x = np.copy(x0)
    obj0 = fun(x)
    optimizer = _Adam(alpha=options['lr'], betas=options['betas'],
                      epsilon=options['epsilon'])

    for i in range(1, options['maxiter']):
        g = jac(x)
        x -= optimizer.step(g, i)
        obj1 = fun(x)
        if np.fabs((obj1 - obj0) / obj1) < options['ftol']:
            break
        obj0 = obj1

    result = {'x': x, 'success': True, 'fun': obj1, 'jac': g, 'nit': i}
    return scipy.optimize.OptimizeResult(result)


class _Adam():
    """Class for Adam.

    Parameters
    ----------
    alpha : float, default 1e-3
    betas : list of float, default [0.9, 0.999]
    epsilon : float, default 1e-8
        There parameters are as defined in the reference [1]_.

    References
    ----------
    .. [1] Diederik P. Kingma, Jimmy Ba, "Adam: A Method for Stochastic
       Optimization", arXiv:1412.6980
    """
    def __init__(self, alpha=1e-3, betas=[0.9, 0.999], epsilon=1e-8):
        self.alpha = alpha
        self.betas = betas
        self.epsilon = epsilon

    def step(self, g, i):
        """Calculate a step of gradient ascent.

        Parameters
        ----------
        g : numpy.ndarray
            Gradient of parameters.
        i : int
            Iteration index.

        Returns
        -------
        numpy.nudarray
            A step at i-th iteration.
        """
        if i == 1:
            self.__m, self.__v = np.zeros_like(g), np.zeros_like(g)

        self.__m = self.betas[0] * self.__m + (1 - self.betas[0]) * g
        self.__v = self.betas[1] * self.__v + (1 - self.betas[1]) * g * g

        hatm = self.__m / (1 - self.betas[0]**i)
        hatv = self.__v / (1 - self.betas[1]**i)
        return self.alpha * hatm / (np.sqrt(hatv) + self.epsilon)




import pickle
import sys
import time
import warnings
from copy import deepcopy

import numpy as np
import scipy.optimize

import sati.objective
import sati.preprocessing
import sati.optimize

class Model:
    """Class for describing a model.

    Parameters
    ----------
    image : ``numpy.ndarray``
        An input image.
    rsp : ``numpy.ndarray``
        An initial guess of the responsibility. This takes a 3D array.
        The first dimension is terrace, and the second and third
        dimensions are positions.
    dist : :class:`sati.distributions.Distribution`
        An initial guess of a distribution.
    poly : :class:`sati.planes.Poly`
        An initial guess of a polynomial plane.
    decay : :class:`sati.planes.Decay`
        An initial guess of a decay plane.
    prior : :class:`sati.distributions.Distribution`, default ``None``
        A prior distribution for ``dist.loc``.
    roi : ``numpy.ndarray``, default ``None``
        A 2D array that has the same shape as the image array and
        specifies an region of interest. Set pixels to be calculated
        to 1.  By default, the whole area is used for the calculation.

    Attributes
    ----------
    objective : ``list`` of :class:`sati.objective.Objective`
        The first element is for the main EM loop, and the second
        element is for the M step. If the quick method is used,
        the second one is not used.
    image : ``numpy.ndarray``
        A copy of the input image.
    subtracted : ``numpy.ndarray``
        An image subtracted by an estimated model plane.
    rsp : ``numpy.ndarray``
        An estimated responsibility.
    dist : :class:`sati.distributions.Distribution`
        An estimated distribution
    poly : :class:`sati.planes.Poly`
        An estimated polynomial plane
    decay : :class:`sati.planes.Decay`
        An estimated decay plane
    prior : :class:`sati.distributions.Distribution`
        An estimated distribution
    n_iter : ``int``
        A number of iterations run by the main EM loop to reach
        the specified tolerance.
    """
    def __init__(self, image, rsp=None, dist=None, poly=None, decay=None,
            prior=None, roi=None):
        self.image, self.rsp, self.dist, self.poly, self.decay, self.prior, \
            self.roi =  image, rsp, dist, poly, decay, prior, roi

    def __validate(self, name, value):
        type_ = self.__ATTRIBUTES[name][0]
        isNoneAllowed = self.__ATTRIBUTES[name][1]
        if isNoneAllowed and (value is None):
            pass
        elif not isinstance(value, type_):
            raise ArgumentTypeError(name, type_)

    @property
    def __ATTRIBUTES(self):
        # The second element of the tuple is if None is allowed.
        return {'image': (np.ndarray, False),
                'rsp': (np.ndarray, True),
                'dist': (sati.distributions.Distribution, True),
                'poly': (sati.planes.Poly, True),
                'decay': (sati.planes.Decay, True),
                'prior': (sati.distributions.Distribution, True),
                'roi': (np.ndarray, True),}

    @property
    def image(self):
        return self.__image

    @image.setter
    def image(self, value):
        self.__validate('image', value)
        self.__image = np.copy(value)

    @property
    def rsp(self):
        return self.__rsp

    @rsp.setter
    def rsp(self, value):
        self.__validate('rsp', value)
        self.__rsp = np.copy(value) if value is not None else None

    @property
    def dist(self):
        return self.__dist

    @dist.setter
    def dist(self, value):
        self.__validate('dist', value)
        self.__dist = deepcopy(value)

    @property
    def poly(self):
        return self.__poly

    @poly.setter
    def poly(self, value):
        self.__validate('poly', value)
        self.__poly = deepcopy(value)

    @property
    def decay(self):
        return self.__decay

    @decay.setter
    def decay(self, value):
        self.__validate('decay', value)
        self.__decay = deepcopy(value)

    @property
    def prior(self):
        return self.__prior

    @prior.setter
    def prior(self, value):
        self.__validate('prior', value)
        self.__prior = deepcopy(value)

    @property
    def roi(self):
        return self.__roi

    @roi.setter
    def roi(self, value):
        self.__validate('roi', value)
        self.__roi = np.copy(value) if value is not None else None

    def __getstate__(self):
        d = {}
        for key in self.__ATTRIBUTES:
            d.update({key: getattr(self, f'_{self.__class__.__name__}__{key}')})
        return d

    def __setstate__(self, state):
        for key in state:
            setattr(self, f'_{self.__class__.__name__}__{key}', state[key])

    def optimize(self, method='auto', maxiter=256, tol=1e-7, verbosity=1,
                 options=None):
        """Optimize model parameters using the expectation-maximization
        (EM) algorithm.

        Parameters
        ----------
        quick : ``bool``, default ``True``
            Use the quick method when possible, that is, Normal or Cauchy
            distribution is used, ``decay`` is ``None``, and ``prior``
            is ``None``.
        method : {'auto', 'l-bfgs-b', 'adam', 'quick'}, default ``auto``
            The method used for the optimization in the M-step.
            When ``auto``, ``quick`` is used if ``dist`` is the normal or
            Cauchy distribution, otherwise ``l-bfgs-b`` is used.
        maxiter: ``int``, default 256
            The maximum number of iterations of the main EM loop.
        tol : ``float``, default 1e-7
            The fractional tolerance to stop the iteration of the main EM loop.
        verbosity : {0, 1, 2}, default 1
            Set 1 to show a summary of result, 2 to show values of
            objective function in the console during calculation, and
            0 to show nothing.

            .. note::

               If you use the Win32 console, 1 (default) or 0 is recommended
               because the values are not shown neatly. (The Win32 console
               does not support the ANSI escape code.)

        options : ``dict``, default ``None``
            A dictionary of optimizer options. If ``None``, the default value
            different for each optimizer is used.

            * For ``method=='l-bfgs-b'``, see `scipy.optimize.minimize(method='L-BFGS-B') <https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html#optimize-minimize-lbfgsb>`_
            * For ``method=='adam'``, see :func:`sati.optimize.minimize`.

        """
        self.__maxiter = maxiter
        self.__verbosity = verbosity
        self.__options = options
        self.objective = sati.objective.Objective(maxiter, tol, verbosity==2)
        self.__method = method.lower()
        self.__m_step = self.__select_m_step()

        self.__setup()
        self.__run_em()
        self.__finalize()

    def __setup(self):
        """Reduce dimension, apply roi (if any), standardize."""
        if self.decay is not None:
            self.__swap_axes()

        self.__t = self.image.reshape(-1)
        if self.rsp is None:
            raise NoAttributeError('rsp')
        self.__gamma = self.rsp.reshape((self.rsp.shape[0], -1))

        if self.roi is not None:
            self.roi = (self.roi > 0).reshape(-1)
            self.__t = np.compress(self.roi, self.__t)
            self.__gamma = np.compress(self.roi, self.__gamma, axis=1)

        self.__t, self.__image_scale = sati.preprocessing.standardize(self.__t)

        if self.poly is None:
            raise NoAttributeError('poly')
        self.poly.setup(self.image.shape, self.roi)

        if self.dist is None:
            raise NoAttributeError('dist')
        self.dist.scaler(self.__image_scale)

        self.__classes = [self.poly, self.dist]

        if self.decay is not None:
            self.decay.setup(self.image.shape, self.__image_scale, self.roi)
            self.__classes.append(self.decay)

        if self.prior is not None:
            self.prior.scaler(self.__image_scale)
            self.__classes.append(self.prior)

    def __swap_axes(self, backward=False):
        self.image = _swap_axes(self.image, self.decay.orgdrct, backward)
        self.rsp = _swap_axes(self.rsp, self.decay.orgdrct, backward)
        if self.roi is not None:
            self.roi = _swap_axes(self.roi, self.decay.orgdrct, backward)
        if backward:
            self.subtracted = _swap_axes(self.subtracted, self.decay.orgdrct,
                                         backward)

    def __select_m_step(self):
        if self.__method not in ['auto', 'l-bfgs-b', 'adam', 'quick']:
            raise ValueError('unknown method.')

        is_quick_available = all([
            hasattr(self.dist, 'update_quick'),
            self.decay is None, self.prior is None])

        if self.__method == 'quick':
            if is_quick_available:
                return self.__m_step_quick
            else:
                raise ValueError("'quick' is not available.")
        if self.__method == 'l-bfgs-b':
            return self.__m_step_ga_bfgs
        if self.__method == 'adam':
            return self.__m_step_ga_adam

        # The following is for 'auto'
        if is_quick_available:
            self.__method = 'quick'
            return self.__m_step_quick
        else:
            self.__method = 'l-bfgs-b'
            return self.__m_step_ga_bfgs

    def __run_em(self):
        start = time.time()
        self.__initial_step()
        try:
            for self.n_iter in range(1, self.__maxiter):
                self.__e_step()
                self.__m_step()
                if (self.objective.isconverged()):
                    break
        except KeyboardInterrupt:
            warnings.simplefilter('ignore', SatiWarning)
            print('\b\bCaught KeyboardInterrupt.')
            pass
        self.__elapsed = time.time() - start

    def __initial_step(self):
        """Do the initial step of the EM loop.

        The initial values of polynomial plane coefficients are determined,
        followed by those of the distribution parameters.
        """
        loc = self.dist.center(self.__t, self.__gamma)
        self.poly.initial_step(self.__t, self.__gamma, loc)

        self.__t_subtracted = self.__t - self.poly.plane
        self.dist.initial_step(self.__t_subtracted, self.__gamma)

        self.objective.append(
            sati.objective.calc(self.dist, self.__t_subtracted,
                                prior=self.prior))

    def __e_step(self):
        """Do the E-step of the EM algorithm and update the gamma."""
        numerator = self.dist.fullpdf(self.__t_subtracted)
        denominator = numerator.sum(axis=0)
        self.__gamma = np.divide(numerator, denominator, where=denominator!=0)

    def __m_step_quick(self):
        """Do the M-step with the quick algorithm."""
        self.dist.update_quick(self.__t_subtracted, self.__gamma)
        obj0 = sati.objective.calc(self.dist, self.__t_subtracted)
        if obj0 < self.objective.value[-1]:
            self.dist.revert()
            obj0 = self.objective.value[-1]

        self.poly.update_quick(self.__t, self.dist.weight, self.dist.loc)
        self.__t_subtracted = self.__t - self.poly.plane
        obj1 = sati.objective.calc(self.dist, self.__t_subtracted)
        if obj1 < obj0:
            self.poly.revert()
            self.__t_subtracted = self.__t - self.poly.plane
            obj1 = obj0

        self.objective.append(obj1)

    def __m_step_ga_bfgs(self):
        """Do the M-step using the L-BFGS-B algorithm."""
        x0 = [c.x0() for c in self.__classes]
        self.__m_step_sizes = [x.size for x in x0]
        bounds = scipy.optimize.Bounds(
                np.concatenate([c.lb() for c in self.__classes], axis=None),
                np.inf)
        result = scipy.optimize.minimize(
                self.__m_step_ga_fun, np.concatenate(x0, axis=None),
                method='L-BFGS-B', jac=self.__m_step_ga_jac, bounds=bounds,
                options=self.__options)

        if not result.success:
            warnings.warn(MstepWarning(result.message))

        self.dist.pi = self.dist.mixing_coef(self.__gamma)
        self.objective.append(result.fun*(-1))

    def __m_step_ga_adam(self):
        """Do the M-step using the ADAM algorithm."""
        x0 = [c.x0() for c in self.__classes]
        self.__m_step_sizes = [x.size for x in x0]
        for c in self.__classes:
            c.retain()

        options = dict(sati.optimize.options_default.items())
        if self.__options is not None:
            options.update(self.__options)

        while True:
            result = sati.optimize.minimize(
                    self.__m_step_ga_fun, np.concatenate(x0, axis=None),
                    method='adam', jac=self.__m_step_ga_jac, options=options)
            # Normal exit
            if result.fun*(-1) > self.objective.value[-1]:
                break
            # Redo the M-step after reducing the learning rate
            options['lr'] *= options['reduction']
            if options['lr'] < np.finfo(float).eps:
                warnings.warn(MstepWarning('Small learning rate'))
                break
            for c in self.__classes:
                c.revert()
            self.__t_subtracted = self.__t - self.poly.plane

        self.dist.pi = self.dist.mixing_coef(self.__gamma)
        self.objective.append(result.fun*(-1))

    def __m_step_ga_update_values(self, x):
        n = self.__m_step_sizes  # shortcut
        self.poly.update_ga(x[:(i:=n[0])])
        self.__t_subtracted = self.__t - self.poly.plane
        self.dist.update_ga(x[i:(i:=i+n[1])])
        if self.decay is not None:
            self.decay.update_ga(x[i:(i:=i+n[2])])
            self.__t_subtracted -= self.decay.plane
        if self.prior is not None:
            self.prior.update_ga(x[i:])

    def __m_step_ga_fun(self, x):
        self.__m_step_ga_update_values(x)
        return sati.objective.calc(self.dist, self.__t_subtracted,
                                   prior=self.prior) * (-1)

    def __m_step_ga_jac(self, x):
        self.__m_step_ga_update_values(x)
        gd0 = self.dist.grad(self.__t_subtracted, self.__gamma)
        gpoly = self.poly.grad(gd0[0])
        gdist = np.stack([x.sum(axis=1) for x in gd0])
        glist = [gpoly, gdist]
        if self.decay is not None:
            gdecay = self.decay.grad(gd0[0])
            glist.append(gdecay)
        if self.prior is not None:
            gprior = self.prior.grad(self.dist.loc, self.__t.size)
            gdist[0] += gprior[0]
            glist.append(gprior[1:])
        return np.concatenate(glist, axis=None) * (-1)

    def __finalize(self):
        n_class = self.__gamma.shape[0]
        if self.roi is not None:
            self.__rsp = np.zeros((n_class, self.image.size))
            np.place(self.__rsp,
                     np.broadcast_to(self.roi, self.__rsp.shape),
                     self.__gamma)
        else:
            self.__rsp = self.__gamma
        self.__rsp.resize((n_class,) + self.image.shape)

        self.dist.scaler(self.__image_scale, backward=True)
        self.poly.finalize(self.__image_scale[1])
        self.subtracted = self.image - self.poly.plane

        if self.decay is not None:
            self.decay.finalize(self.__image_scale[1])
            self.subtracted -= self.decay.plane
            self.__swap_axes(backward=True)

        if self.prior is not None:
            self.prior.scaler(self.__image_scale, backward=True)

        # show messages
        if not self.objective.isconverged():
            warnings.warn(NotConvergedWarning(self.__maxiter))

        if self.__verbosity > 0:
            print(f'method: {self.__method}')
            print(self.objective)
            print(self.dist)
            if self.decay is not None:
                print(self.decay)
            if self.prior is not None:
                print(self.prior)
            print(_elapsed(self.__elapsed))

    def pickle(self, file):
        """Pickle a model.

        This is a wrapper of the following.

        .. code-block::

            with open(file, 'wb') as f:
                pickle.dump(self, f)

        Parameters
        ----------
        file : str
            A path-like object of the file to be saved.
        """
        with open(file, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def unpickle(file):
        """Unpickle a model.

        This is a wrapper of the following.

        .. code-block::

            with open(file, 'rb') as f:
                return pickle.load(f)

        Parameters
        ----------
        file : str
            A path-like object of the file to be opened.
        """
        with open(file, 'rb') as f:
            return pickle.load(f)


def _swap_axes(array, orgdrct, backward=False):
    """By flipping and swapping axes, reorder data array to the
    standard order.

    Forward (backward==False) is from real to standardized,
    backward (backward==True) is from standardized to real.
    Since flip and transpose are not commutable, swap must come at
    the last for forward, and at the beginning for backward. """
    if 'y' in orgdrct and backward:
        array = np.swapaxes(array, array.ndim - 2, array.ndim - 1)
    if 'r' in orgdrct:
        array = np.flip(array, array.ndim - 1)
    if 't' in orgdrct:
        array = np.flip(array, array.ndim - 2)
    if 'y' in orgdrct and not backward:
        array = np.swapaxes(array, array.ndim - 2, array.ndim - 1)
    return array


def _elapsed(elapsed):
    h, m, s = int(elapsed // 3600), int(elapsed // 60 % 60), int(elapsed % 60)
    if elapsed < 120:
        return f'elapsed: {elapsed:.3f} s'
    elif elapsed < 3600:
        return f'elapsed: {m} m {s} s'
    else:
        return f'elapsed: {h} h {m} m {s} s'


class SatiError(Exception):
    """A base class of exceptions thrown by sati.Model."""
    pass


class ArgumentTypeError(SatiError):
    """An exception thrown when type of an argument is wrong."""
    def __init__(self, name, type_):
        self.__name = name
        self.__type = type_

    def __str__(self):
        return f"'{self.__name}' must be an instance of " \
               f'{self.__type.__module__}.{self.__type.__name__}'


class NoAttributeError(SatiError):
    """An exception thrown when an attribute is None."""
    def __init__(self, attr):
        self.__attr = attr

    def __str__(self):
        return f"'{self.__attr}' must be given."


class SatiWarning(Warning):
    """A base class of warning thrown by sati.Model."""
    pass


class NotConvergedWarning(SatiWarning):
    """A warning thrown when the optimization did not reach to the
    specfied tolerance. """
    def __init__(self, n_iter):
        self.__n_iter = n_iter

    def __str__(self):
        return f'Not converged after {self.__n_iter} iterations.'


class MstepWarning(SatiWarning):
    """A warning thrown when the optimizer of M-step fails."""
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return f'The optimizer of M-step exited unsuccessfully.\n' \
               f'{self.__msg}'




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sati']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.1,<4.0.0',
 'numexpr>=2.7.1,<3.0.0',
 'numpy>=1.18.0,<2.0.0',
 'scipy>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'sati',
    'version': '1.0.0',
    'description': 'Statistical analysis of topographic image',
    'long_description': 'Sati\n====\nStatistical analysis of Topographic Image\n\nThis package provides\n\n* subtracting background (a polynomial surface, logarithmic decays, exponential decays)\n* labeling terraces\n* estimating terrace heights\n* estimating a unit height of steps\n\nThe subtraction is possible even in the presence of steps.\n\nInstall\n-------\nYou can install the package from the git repogitory using ``pip``\n::\n\n  $ pip install sati\n\nRequirements\n------------\nSati requires the following dependencies:\n\n* python 3.8\n* matplotlib\n* numexpr\n* numpy\n* scipy\n\nDocuments\n---------\nThe usage examples and the API documentation are available at https://yuksk.github.io/sati/index.html\n\n\n',
    'author': 'Yuhki Kohsaka',
    'author_email': 'yuhki.kohsaka@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yuksk/sati/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sys

import numpy as np


def calc(dist, t_subtracted, prior=None):
    """Calculate a value of objective function.

    Parameters
    ----------
    dist : a subclass of :class:`sati.distributions.Distribution`
        A distribution used in a model.
    t_subtracted : ``numpy.ndarray``
        Subtracted one-dimensionalized image.
    prior : :class:`sati.distributions.VonMises`
        A prior distribution.

    Returns
    -------
    float :
        A value of objective function.
    """
    likelihood_at_pixel = dist.fullpdf(t_subtracted).sum(axis=0)
    np.place(likelihood_at_pixel, likelihood_at_pixel==0, sys.float_info.min)
    if prior is None:
        return np.log(likelihood_at_pixel).sum()
    else:
        return np.log(likelihood_at_pixel).sum() \
                + prior.loglikelihood_at_pixel(dist.loc) * t_subtracted.size


class Objective():
    """Class for an objective function.

    Parameters
    ----------
    maxiter : ``int``
        The maximum number of iterations of the main EM loop.
    tol : ``float``
        The fractional tolerance to stop the iteration.
    verbose : ``bool``
        If ``True``, show values of objective function in the console
        during calculation.

    Attributes
    ----------
    value, diff : ``list`` of ``float``
        Objective function and its relative difference. Values are
        appended at each cycle of iterations.
    """
    def __init__(self, maxiter, tol, verbose):
        self.__tol, self.__verbose = tol, verbose
        self.value = []
        self.diff = []

        # used to show value and diff.
        self.__precision = (6, 4)

        # values used for the console update.
        self.__lines = 7
        self.__index_width = int(np.trunc(np.log10(maxiter))) + 1

    def __str__(self):
        return f'iterations: {len(self.value)-1}\n' \
               f'objective: {self.value[-1]:.{self.__precision[0]}e}'

    def append(self, value):
        """Append a value to the list of values and update list of differences.

        Parameters
        ----------
        value : float
            A value of an objective function.
        """
        if len(self.diff) == 0 or value == 0:
            self.diff.append(np.nan)
        else:
            self.diff.append(np.fabs((value - self.value[-1]) / value))
        self.value.append(value)

        if self.__verbose:
            self.__update_console()

    def isconverged(self):
        return self.diff[-1] < self.__tol

    def __update_console(self):
        def oneline(i):
            """Construct a line."""
            linestr = ''
            # If the objective decreases, change the character color.
            if i > 0 and self.value[i] < self.value[i - 1]:
                linestr += '\033[32m'  # green
            # Content of line
            linestr += ' '.join([
                f"{i:<{self.__index_width}}",
                f"{self.value[i]:+.{self.__precision[0]}e}",
                f"({self.diff[i]:.{self.__precision[1]}e})"])
            # Revert the character color.
            linestr += '\033[39m'
            return linestr

        if len(self.value) == 1:
            # Prepare lines in case the prompt is close to the bottom.
            print('\n' * (self.__lines- 1))

        print(f"\033[{self.__lines}F", end='')
        L = len(self.value)
        nlines = min(L, self.__lines)
        print('\n'.join([oneline(i) for i in range(L - nlines, L)]))
        print('\n' * (self.__lines - nlines), end='')

